
import email
import smtplib
import ssl

from email import encoders
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from .helpers import *

ADMINS = []
USERS = []


class EmailHandler:
    '''
    Sends emails, with an easier to use interface

    send an isolated email with the class method EmailHandler.send_one(*args) [NOT IMPLEMENTED]


HOW TO
    Initialize like so:

            email_handler = EmailHandler(server (ex: 'smtp.gmail.com'),
                                            default_from (what the default sending email is),
                                            password (optional if server == 'localhost' else required),
                                            port (optional; default == 465),
                                            admins (optional; if you want to specify common admins to send mail to),
                                            users (optional; if you want to specify common users to send mail to)

HOW TO
    Send email, after initialized, like such:

            email_handler.send_mail('my subject', 'my body', img_attachments=images)
    '''

    def __init__(
            self, server, default_from, password=None, port=None,
            admins=ADMINS, users=USERS):
        self.port = port or 465
        self.password = password if server != 'localhost' else None
        self.default_from = default_from
        self.server_name = server
        self.admins = admins
        self.users = users

    @classmethod
    def init_from_test_settings(cls):
        '''
Instantiates this object ready to send mail to test server

        Run this command in a separate terminal to initialize testing server:
                $ python -m smtpd -n -c DebuggingServer localhost:1025
        '''
        return cls('localhost', port=1025, default_from='testing@example.com')

    @classmethod
    def init_default_settings(cls):

        return cls('smtp.gmail.com', '<your gmail email here>',
                   '<token or password for email here>', 465)

    @classmethod
    def send_one():
        '''
        needs further implementation in a 'best way undetermined' kind of way
        '''

        server = input('What is the server you are sending from? ')
        from_username = input(
            'What user email are you sending from? (ie: bob@gmail.com) ')
        from_password = input(f"Password for {from_username}? ")

    def initialize_server(self):
        '''
        Opens connection to the server as defined by self.server_name
        '''
        if self.server_name == 'localhost':
            self.server = smtplib.SMTP(self.server_name, self.port)
        else:
            self.server = smtplib.SMTP_SSL(self.server_name, self.port)

    def send_mail(self, subject, body, to=None, from_=None,
                  img_attachments=[], img_types='email format'):
        '''
        sends emails to list of users defined by <to> user_form

        to: list (defaults to self.admins)
                list of users to send email to

        img_attachments: list of images
                expects email ready format ( see ImageHandler.get_email_ready_format() ); two tuple (image_filename, image_bytestring)

        img_types: str
                sets type of element in img_attachments; default == 'email format'
                also supports 'bytestring' if you are passing a list of bytestrings as img_attachments; names will be 'image1.png', 'image2.png', etc.

        #TODO add support for other file attachment types
        #TODO? add support for other image types
        '''
        if to == None:
            to = self.admins

        if from_ == None:
            from_ = self.default_from

        self.initialize_server()

        if self.server._host != 'localhost':
            from_ = self.login_to_server(from_)

        msg = MIMEMultipart()
        msg["From"] = from_
        msg["To"] = ', '.join(to)
        msg["Subject"] = subject

        msg.attach(MIMEText(body))

        if img_attachments:
            msg = self.attach_images(msg, img_attachments, img_types)

        self.server.sendmail(from_, to, msg.as_string())
        del msg

        self.close_conn()

    def login_to_server(self, from_):
        '''
        handle setting password (if necessary) and logging in with validation of success or authentication error
        '''
        password = self.set_password()
        while True:
            try:
                self.server.login(from_, password)
                break
            except smtplib.SMTPAuthenticationError:
                option = input(
                    'There was an error with authentication. Please press 1 to edit username or 2 to edit password: ')
                if option == '1':
                    from_ = input('Username: ')
                elif option == '2':
                    password = input('Password: ')

        return from_

    def set_password(self):
        '''
        sets password and returns it
        '''
        if self.password:
            password = self.password
        else:
            password = input(
                "Please input the password associated with the sending email: ")

        return password

    def attach_images(self, msg, img_attachments, img_types='email format'):
        '''
        attaches images based on the specific format held by <img_attachments> specified by <img_types>
        '''
        if img_types == 'email format':
            for img in img_attachments:
                image = MIMEImage(img[1], name=img[0])
                msg.attach(image)
        elif img_types == 'bytestring':
            for idx, img in enumerate(img_attachments):
                image = MIMEImage(img, name=f"image{idx}")
                msg.attach(image)

        return msg

    def close_conn(self):
        self.server.quit()



import os
import os.path
import json
from .helpers import *
import re
import pickle
from .decorators.files import get_filename
from .decorators.logging import log_error


class FileOperations:

    '''
    Working directory defaults to the current working directory

    If sub_directory == True, then the write folder is found as a subdirectory of the current working directory

    If start_home == True, then start from the home directory in order to find the working_directory
        (this clearly negates sub_directory, however sub_directory has prio if True is passed for both)

    Working directory can be set again with self.set_working_directory(*args)

    Methods will default to writing to the default_filename if their filename is not declared. Same for directory
    '''

    def __init__(self, working_directory='', sub_directory=False,
                 start_home=False, default_filename=None):

        self.set_working_directory(
            working_directory, sub_directory, start_home)

        self.default_filename = default_filename

        if self.default_filename != None:
            self.default_filename_with_path = self.working_directory + '/' + self.default_filename

            # try to create file so we dont get any errors
            try:
                open(self.default_filename_with_path, 'x')
            except:
                pass

        # self.create directory if dir doesnt exist - this is a todo item, naming not immediately evident

    def set_working_directory(
            self, working_directory='', sub_directory=False, start_home=False):

        if not working_directory.startswith('/') and working_directory:
            working_directory = '/' + working_directory

        if sub_directory:
            self.working_directory = os.getcwd() + working_directory
        elif start_home:
            self.working_directory = os.environ.get('HOME') + working_directory
        else:
            self.working_directory = working_directory or os.getcwd()

    # @get_filename('.json')
    def write_json(self, data, filename=None):
        '''
        writes json to directory <self.working_directory> with name <filename>
        '''
        with open(f"{self.working_directory}/{filename}", 'w') as outfile:
            json.dump(data, outfile)

    # @get_filename('.pickle')

    def write_pickle(self, data, filename=None):
        if not filename:
            filename = self.default_filename
        with open(f"{self.working_directory}/{filename}", "wb") as file_:
            pickle.dump(data, file_)

    def read_pickle(self, filename=None):
        if not filename:
            filename = self.default_filename
        with open(f"{self.working_directory}/{filename}", "rb") as file_:
            output = pickle.load(file_)
        return output

    # @get_filename('.txt')
    def write_text(self, data, filename, method='a+'):
        with open(f"{self.working_directory}/{ filename }", method) as file_:
            file_.write(data)

    # @get_filename('.txt')
    @log_error('file_operations.log', True)
    def read_text(self, filename):
        with open(f"{self.working_directory}/{ filename }", 'r') as file_:
            file.seek(0)
        for line in file_:
            print(line.strip())

    def get_file_parts(self, file_):
        '''
        file needs appropriate path headed

        returns filename, file_extension
        '''
        return os.path.splitext(file_)

    def delete_file(self, file_):
        try:
            os.remove(f"{self.working_directory}/{file_}")
        except:
            print('No file to delete')

    def append_text(self, text, file_with_path=None):

        pass

    def verify_defaults(self):

        pass

    def get_files_in_directory(
            self, directory=None, recurse=False, full_path=False):
        '''
        recursion not handled yet
        '''

        directory = directory or self.working_directory

        files = []

        for root, directories, filenames in os.walk(directory):
            for filename in filenames:
                if full_path == False:
                    file_ = filename
                else:
                    file_ = os.path.join(root, filename)

                files.append(file_)

        return files

    def regex_sub_files(self, regex, files, sub='', count=1):
        '''
        calls a regex sub for each file in files

        TODO: make it so that if sub == 'index' or 'idx' (not sure which), then we can substitute in the index of the file in files

        includes check to make sure returned filename is acceptable and requires input
        '''
        pattern = re.compile(regex)
        new_filenames = []

        for idx, file_ in enumerate(files):

            new_filename = pattern.sub(
                sub, self.get_file_parts(file_)[0],
                count=count)

            if idx == 0:
                while True:
                    ans = input(
                        f"{new_filename} is your new filename, is this acceptable? (y/n)")
                    if ans.lower() == 'y':
                        break
                    elif ans.lower() == 'n':
                        print('Sorry to hear your regex failed, stupid dum dum.')
                        return
                    else:
                        pass
            new_filenames.append(new_filename + self.get_file_parts(file_)[1])
            os.rename(
                f"{self.working_directory}/{file_}",
                f"{self.working_directory}/{new_filename + self.get_file_parts(file_)[1]}")

        print('Your new filenames: \n')
        for file_ in new_filenames:
            print(file_)



from tabulate import tabulate


# Create an html table easily with tabulate like so:

# from tabulate import tabulate
# tabulate(<your_data>, headers=<your_headers>, tablefmt="html")
# where <your_data> is of the form [[row1 data], [row2 data], ....etc]
# where <your_headers> is of the form [header1, header2, header3, ....etc]

# this will output clean html table
# helper for this defined below

# all supported table formats by tabulate
# "plain"
# "simple"
# "github"
# "grid"
# "fancy_grid"
# "pipe"
# "orgtbl"
# "jira"
# "presto"
# "pretty"
# "psql"
# "rst"
# "mediawiki"
# "moinmoin"
# "youtrack"
# "html"
# "latex"
# "latex_raw"
# "latex_booktabs"
# "textile"


def create_table_from_objects(
        objects, keys_as_headers, format='fancy_grid', **kwargs):
    '''
    <objects> is a list of python dicts
    <keys_as_headers> are the desired keys from the objects that you wish to display values for
    <format> is simply passed to tabulate (see above for all supported formats)
        common formats are : "html", "fancy_grid"

    pass any required optional kwargs to the tabulate function

    data is composed of rows so the naming is such

    return value is table in specified format

    if sorting is desired then it is probably easier to simply view in browser (local or not) by using Datatables
        because sorting can be applied to any field
    '''

    data = []
    for obj in objects:
        row = []
        for key in keys_as_headers:
            row.append(obj[key])
        data.append(row)

    headers = keys_as_headers

    return tabulate(data, headers=headers, tablefmt=format, **kwargs)
